﻿using MVR.FileManagementSecure;
using SimpleJSON;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.InteropServices;

namespace JayJayWon
{
    public class PluginManifest
    {
        private Dictionary<string, PluginManifestEntry> pluginManinfestDict;

        private HashSet<string> creatorsSet;
        private Dictionary<string, bool> creatorsPreferredDict;
        private Dictionary<string, bool> creatorsConcealedDict;

        private TagsManifest tagsManifest;

        public PluginManifest(TagsManifest _tagsManifest)
        {
            pluginManinfestDict = new Dictionary<string, PluginManifestEntry>();
            creatorsPreferredDict = new Dictionary<string, bool>();
            creatorsConcealedDict = new Dictionary<string, bool>();
            creatorsSet = new HashSet<string>();
            tagsManifest = _tagsManifest;

            Initialise();
        }

        public PluginManifestEntry GetPluginManifestEntry(string pmeName)
        {
            if (pluginManinfestDict.ContainsKey(pmeName)) return (pluginManinfestDict[pmeName]);
            return null;
        }

        private void Initialise()
        {
            pluginManinfestDict.Clear();
            HashSet<string> varFiles = Utils.GetFilesAtPathRecursive("AddonPackages", "*.var");
            foreach (string varFile in varFiles)
            {
                ProcessNewVARFile(varFile);
            }
            ProcessLocalScriptFiles();
        }
        public void RefreshVARPluginManifest()
        {
            ClearManifestDict(false);
            HashSet<string> varFiles = Utils.GetFilesAtPathRecursive("AddonPackages", "*.var");
            foreach (string varFile in varFiles)
            {
                ProcessNewVARFile(varFile);
            }
        }
        public void RefreshLocalPluginManifest()
        {
            ClearManifestDict(true);
            ProcessLocalScriptFiles();
        }


        private void ClearManifestDict(bool keepVARFilesOnly)
        {
            List<string> keysToClear = new List<string>();
            foreach (KeyValuePair<string,PluginManifestEntry> kvp in pluginManinfestDict)
            {
                if (kvp.Value.isLocal == keepVARFilesOnly) keysToClear.Add(kvp.Key);
            }
            foreach (string key in keysToClear) pluginManinfestDict.Remove(key);
        }

        private void StoreCreatorMetaData(string creator)
        {
            if (creatorsSet.Contains(creator))
            {
                JSONClass metaData = new JSONClass();
                metaData["PAversion"] = Const.PAversion;
                metaData["PAversionFloat"].AsFloat = Const.PAversionFloat;
                metaData["metaDataStoreFormat"].AsInt = Const.MetaDataStoreFormat;
                metaData["creatorName"] = creator;

                metaData["creatorConcealed"].AsBool = creatorsConcealedDict[creator];
                metaData["creatorPreferred"].AsBool = creatorsPreferredDict[creator];

                SuperController.singleton.SaveJSON(metaData, Const.CreatorMetaDataFolder + "\\" + creator + Const.CreatorMetaDataFileExt);
            }
        }

        private void RestoreCreatorMetaData(string creator)
        {
            if (FileManagerSecure.FileExists(Const.CreatorMetaDataFolder + "\\" + creator + Const.CreatorMetaDataFileExt))
            {
                JSONClass metaData = (JSONClass) SuperController.singleton.LoadJSON(Const.CreatorMetaDataFolder + "\\" + creator + Const.CreatorMetaDataFileExt);
                if (metaData["metaDataStoreFormat"] != null && metaData["metaDataStoreFormat"].AsInt > Const.MetaDataStoreFormat) Utils.LogDebug("<RestoreCreatorMetaData> Creator Metadata Store Format for '"+ Const.CreatorMetaDataFolder + "\\" + creator + Const.CreatorMetaDataFileExt+"' is a higher version than supported.");

                if (metaData["creatorConcealed"] !=null) creatorsConcealedDict[creator] = metaData["creatorConcealed"].AsBool;
                if (metaData["creatorPreferred"] != null) creatorsPreferredDict[creator] = metaData["creatorPreferred"].AsBool;
            }
        }

        public bool GetCreatorPreferred (string creatorName)
        {
            if (creatorsPreferredDict.ContainsKey(creatorName)) return (creatorsPreferredDict[creatorName]);
            else Utils.LogDebug("GetCreatorPreferred: Creator Name '"+creatorName+"' not found");
            return (false);
        }

        public void SetCreatorPreferred(string creatorName, bool value)
        {
            if (creatorsPreferredDict.ContainsKey(creatorName)) creatorsPreferredDict[creatorName]=value;
            else Utils.LogDebug("SetCreatorPreferred: Creator Name '" + creatorName + "' not found");
            StoreCreatorMetaData(creatorName);
        }

        public bool GetCreatorConcealed(string creatorName)
        {
            if (creatorsConcealedDict.ContainsKey(creatorName)) return (creatorsConcealedDict[creatorName]);
            else Utils.LogDebug("GetCreatorConcealed: Creator Name '" + creatorName + "' not found");
            return (false);
        }

        public void SetCreatorConcealed(string creatorName, bool value)
        {
            if (creatorsConcealedDict.ContainsKey(creatorName)) creatorsConcealedDict[creatorName] = value;
            else Utils.LogDebug("SetCreatorConcealed: Creator Name '" + creatorName + "' not found");
            StoreCreatorMetaData(creatorName);
        }

        public List<string> GetCreatorsList( bool includeConcealed=false, int preferredFilter = PreferredFilter.any, bool excludeLocal = true)
        {
            List<string> creatorsList = new List<string>();
            creatorsList = creatorsSet.Where(x => {
                if ((includeConcealed || !creatorsConcealedDict[x]) &&
                (preferredFilter== PreferredFilter.any || (preferredFilter ==PreferredFilter.preferred && creatorsPreferredDict[x])|| (preferredFilter==PreferredFilter.notPreferred && !creatorsPreferredDict[x]))) return true;
                else return false;
            }).ToList<string>();

            if (excludeLocal) creatorsList.Remove("LOCAL");
            creatorsList.Sort();
            return (creatorsList);
        }

        public List<string> GetPluginsList(bool allCreators = true, bool localCreators = false, string creatorName = "", bool displayNames = false, int preferredFilter = PreferredFilter.any, bool showConcealedCreators = false, bool showConcealedPlugins =false, int pluginTypeFilter=0, JSONStorableFixedIndexStringChoooser tagsFilterJSON=null)
        {
            List<string> pluginsList = new List<string>();

            bool anyTagFilter = tagsFilterJSON.isValFixedIndex;
            PluginTag filteredTag = tagsManifest.GetTag(tagsFilterJSON.displayVal);

            pluginsList = pluginManinfestDict.Values
                .Where(x =>
                {
                    if (((allCreators && (!x.isLocal || localCreators)) || (localCreators && x.isLocal) || (!allCreators && !localCreators && x.creator == creatorName )) &&
                        !x.referencedByCSList &&
                        ((x.anyVariantsPreferred && preferredFilter==PreferredFilter.preferred)||(!x.anyVariantsPreferred && preferredFilter==PreferredFilter.notPreferred)||preferredFilter==PreferredFilter.any) &&
                        (pluginTypeFilter==0 || x.pluginType== pluginTypeFilter) &&
                        (showConcealedPlugins || !x.allVariantsConcealed)&&
                        (anyTagFilter || x.pluginTagSet.IsTagEnabled(filteredTag)) &&
                        (showConcealedCreators || !creatorsConcealedDict[x.creator])) return true;
                    else return false;
                })
                .Select(x =>
                {
                    if (displayNames)
                    {
                        if ((x.isLocal && !allCreators) || x.pluginNameNoFileExt.ToLower() == x.varPackageName.ToLower()) return x.pluginNameNoFileExt;
                        else if (x.isLocal) return x.pluginNameNoFileExt + " [" + x.creator + "]";
                        else return x.pluginNameNoFileExt + " [" + x.varPackageName + "]";
                    }
                    else return (x.pluginManifestEntryName);
                }).ToList<string>();

            return (pluginsList);
        }
        public List<string> GetPluginTypesList(JSONStorableCreatorChooser creatorJSON, bool typeEnum = false)
        {
            List<string> list = new List<string>();
            foreach(int i in PluginAtomTypes.GetAllTypeEnums())
            {
                int typeCount = pluginManinfestDict.Values.Count(p => p.pluginType==i && p.CheckCreatorMatchesFilter(creatorJSON));
                if (typeCount > 0)
                {
                    if (typeEnum) list.Add(i.ToString());
                    else list.Add(PluginAtomTypes.Label(i));
                }
            }
            return (list);
        }

        public List<string> GetPluginTagsList(JSONStorableCreatorChooser creatorJSON, JSONStorableStringChooser pluginTypeJSON)
        {
            List<string> list = new List<string>();
            foreach (PluginTag pluginTag in tagsManifest.pluginTags)
            {
                int tagCount = pluginManinfestDict.Values.Count(p => (int.Parse(pluginTypeJSON.val) == PluginAtomTypes.All || p.pluginType == int.Parse(pluginTypeJSON.val)) && p.CheckCreatorMatchesFilter(creatorJSON)&& p.pluginTagSet.IsTagEnabled(pluginTag));
                if (tagCount > 0)
                {
                    list.Add(pluginTag.tagLabel);
                }
            }
            return (list);
        }

        private void AddCreatorToManifest(string creator)
        {
            if (!creatorsSet.Contains(creator))
            {
                creatorsSet.Add(creator);
                if (!creatorsConcealedDict.ContainsKey(creator)) creatorsConcealedDict.Add(creator, false);
                if (!creatorsPreferredDict.ContainsKey(creator)) creatorsPreferredDict.Add(creator, false);
                if (creator != "LOCAL") RestoreCreatorMetaData(creator);
            }

        }
        private void RemovePluginManifestForAuthor(PluginManifestEntry pme, string creator)
        {

            // TBD - Check there are no PMEs for Creator left and remove Creator if true?
        }

        private void CreateManifestEntriesForCSFiles(HashSet<string> csFiles, VARPackageID varPackageID)
        {
            foreach (string csFile in csFiles)
            {
                string pluginName = FileManagerSecure.GetFileName(csFile);
                string pluginPath = FileManagerSecure.GetDirectoryName(csFile, true);
                string pluginManifestEntryName = "LOCAL" + "." + pluginName;

                if (varPackageID!=null) pluginManifestEntryName = varPackageID.varCreator + "." + varPackageID.varPackageName + "." + pluginName;

                if (pluginManinfestDict.ContainsKey(pluginManifestEntryName))
                {
                    if (varPackageID!=null) pluginManinfestDict[pluginManifestEntryName].AddVARVersion(varPackageID.varVersion, pluginPath);
                    else pluginManinfestDict[pluginManifestEntryName].AddVARVersion(0, pluginPath);
                }
                else
                {
                    if (varPackageID != null)
                    {
                        PluginManifestEntry pme = new PluginManifestEntry(varPackageID.varCreator, pluginPath, pluginName, varPackageID.varPackageName, varPackageID.varVersion, tagsManifest, false);
                        pluginManinfestDict.Add(pluginManifestEntryName, pme);
                        AddCreatorToManifest(varPackageID.varCreator);
                        
                    }
                    else
                    {
                        PluginManifestEntry pme = new PluginManifestEntry(pluginPath, pluginName, tagsManifest, false);
                        pluginManinfestDict.Add(pluginManifestEntryName, pme);
                        AddCreatorToManifest("LOCAL");
                    }
                        
                }

                if (csFile.EndsWith(".cslist"))
                {
                    string cslFilePath = csFile;
                    if (varPackageID != null) cslFilePath = varPackageID.varFileName + ":/" + csFile;
                    if (FileManagerSecure.FileExists(cslFilePath))
                    {
                        string[] stringSeparators = new string[] { ".cs", "\n", "\r\n" };

                        string csListFileContents = FileManagerSecure.ReadAllText(cslFilePath);
                        string[] csListPluginFiles = csListFileContents.Split(stringSeparators, System.StringSplitOptions.RemoveEmptyEntries);

                        foreach (string csFileRef in csListPluginFiles)
                        {
                            if (!(String.IsNullOrEmpty(csFileRef) || csFileRef.Trim().Length == 0))
                            {
                                string csFilePathRef; ;
                                if (varPackageID != null)
                                {
                                    if (!csFileRef.Contains(':')) csFilePathRef = varPackageID.varFileName + ":/" + pluginPath + "/" + csFileRef + ".cs";
                                    else csFilePathRef = csFileRef + ".cs";
                                }
                                else csFilePathRef = pluginPath + "/" + csFileRef + ".cs";

                                string normalizedFileRef = FileManagerSecure.NormalizePath(FileManagerSecure.GetFullPath(csFilePathRef));

                                if (FileManagerSecure.FileExists(normalizedFileRef))
                                {
                                    string csPluginName = FileManagerSecure.GetFileName(normalizedFileRef);
                                    string csFilePluginManifestEntryName = "LOCAL:" + csPluginName;
                                    if (varPackageID != null) csFilePluginManifestEntryName = varPackageID.varCreator + "." + varPackageID.varPackageName + "." + csPluginName;

                                    if (pluginManinfestDict.ContainsKey(csFilePluginManifestEntryName))
                                    {
                                        if (varPackageID != null) pluginManinfestDict[csFilePluginManifestEntryName].SetVARVersionReferencedInCSList(varPackageID.varVersion);
                                        else
                                        {
                                            RemovePluginManifestForAuthor(pluginManinfestDict[csFilePluginManifestEntryName], "LOCAL");
                                            pluginManinfestDict.Remove(csFilePluginManifestEntryName);

                                        }
                                    }
                                }
                                else if (varPackageID != null)
                                {
                                    Utils.LogDebug("<CreateManifestEntriesForCSFiles> '" + normalizedFileRef + "' referenced in '" + cslFilePath + "' does not exist. Possible malformed cslist file.");
                                }
                            }



                        }
                    }
                    else {
                        if (varPackageID!=null) Utils.LogDebug("<CreateManifestEntriesForCSFiles> " + cslFilePath + " referenced in 'meta.json' does not exist. Possible mispackaged VAR.");
                        else Utils.LogDebug("<CreateManifestEntriesForCSFiles> " + cslFilePath + " does not exist.");
                    }
                }
            }
        }
        private void ProcessNewVARFile(string varFilePath)
        {
            if (FileManagerSecure.FileExists(varFilePath + ":\\meta.json"))
            {
                JSONClass varMetaFile = (JSONClass)SuperController.singleton.LoadJSON(varFilePath + ":\\meta.json");

                JSONArray contentList = null;
                if (varMetaFile!=null && varMetaFile["contentList"]!=null) contentList= varMetaFile["contentList"].AsArray;
                else
                {
                    Utils.LogDebug("<ProcessNewVARFile>: Possible malformed meta.json file in VAR '" + varFilePath + "'");
                    return;
                }

                HashSet<string> csFiles = new HashSet<string>();
                HashSet<string> csListFiles = new HashSet<string>();

                // Loops through each of the entries in the Content List section of the meta.json file - identifying all .cs and .cslist files in the VAR
                foreach (JSONNode node in contentList)
                {
                    string fileName = FileManagerSecure.NormalizePath(node);
                    if (fileName.StartsWith("Custom/Scripts"))
                    {
                        if (fileName.EndsWith(".cs") || fileName.EndsWith(".dll")) csFiles.Add(fileName);
                        else if (fileName.EndsWith(".cslist")) csListFiles.Add(fileName);
                        else if (FileManagerSecure.IsDirectoryInPackage(varFilePath + ":\\" + fileName))
                        {
                            //Where the contents list references a folder then recursively search the folder for .cs and .cslist files
                            HashSet<string> orphanCsFiles = Utils.GetFilesAtPathRecursive(varFilePath + ":\\" + fileName, "*.cs", true);
                            csFiles.UnionWith(orphanCsFiles);
                            HashSet<string> orphanDllFiles = Utils.GetFilesAtPathRecursive(varFilePath + ":\\" + fileName, "*.dll", true);
                            csFiles.UnionWith(orphanDllFiles);
                            HashSet<string> orphanCsListFiles = Utils.GetFilesAtPathRecursive(varFilePath + ":\\" + fileName, "*.cslist", true);
                            csListFiles.UnionWith(orphanCsListFiles);
                        }
                    }
                }
                if (csFiles.Count > 0 || csListFiles.Count > 0)
                {
                    VARPackageID varPackageID = new VARPackageID(varFilePath);
                    if (varPackageID.varCreator == "AcidBubbles" && varPackageID.varPackageName == "Timeline" && csListFiles.Count == 0)
                    {
                        string temp = "Custom/Scripts/AcidBubbles/Timeline/VamTimeline.AtomAnimation.cslist";
                        if (FileManagerSecure.FileExists(varPackageID.varFileName + ":/" + temp)) csListFiles.Add(temp);

                        temp = "Custom/Scripts/AcidBubbles/Timeline/VamTimeline.Controller.cslist";
                        if (FileManagerSecure.FileExists(varPackageID.varFileName + ":/" + temp)) csListFiles.Add(temp);

                        temp = "Custom/Scripts/VamTimeline.AtomAnimation.cslist";
                        if (FileManagerSecure.FileExists(varPackageID.varFileName + ":/" + temp)) csListFiles.Add(temp);

                        temp = "Custom/Scripts/VamTimeline.Controller.cslist";
                        if (FileManagerSecure.FileExists(varPackageID.varFileName + ":/" + temp)) csListFiles.Add(temp);
                    }
                    if (varPackageID.varCreator == "AcidBubbles" && varPackageID.varPackageName == "ColliderEditor" && csListFiles.Count == 0)
                    {
                        string temp = "Custom/Scripts/AcidBubbles/ColliderEditor/ColliderEditor.cslist";
                        if (FileManagerSecure.FileExists(varPackageID.varFileName + ":/" + temp)) csListFiles.Add(temp);
                    }
                    if (varPackageID.varCreator == "geesp0t" && varPackageID.varPackageName == "EasyMoan" && csListFiles.Count == 0)
                    {
                        string temp = "Custom/Scripts/geesp0t/EasyMoan/EasyMoan.cslist";
                        if (FileManagerSecure.FileExists(varPackageID.varFileName + ":/" + temp)) csListFiles.Add(temp);
                    }
                    if (varPackageID.varCreator == "juniperD" && varPackageID.varPackageName == "Catalog3" && csFiles.Count > 0)
                    {
                        string csFileRemove="";
                        foreach (string csf in csFiles) if (csf.StartsWith("Custom/Scripts/Catalog3/obj/Debug/.NET")) csFileRemove = csf;
                        if (csFileRemove!="") csFiles.Remove(csFileRemove);
                    }
                    CreateManifestEntriesForCSFiles(csFiles, varPackageID);
                    CreateManifestEntriesForCSFiles(csListFiles, varPackageID);
                }
            }
        }
        private void ProcessLocalScriptFiles()
        {
            HashSet<string> localCSFiles = Utils.GetFilesAtPathRecursive("Custom/Scripts", "*.cs");
            HashSet<string> localDllFiles = Utils.GetFilesAtPathRecursive("Custom/Scripts", "*.dll");
            localCSFiles.UnionWith(localDllFiles);
            CreateManifestEntriesForCSFiles(localCSFiles, null);
            HashSet<string> localCSListFiles = Utils.GetFilesAtPathRecursive("Custom/Scripts", "*.cslist");
            CreateManifestEntriesForCSFiles(localCSListFiles, null);
        }
    }

    public class PluginManifestEntry
    {
        public string creator { get; private set; }
        public string varPackageName { get; private set; }
        public string pluginName { get; private set; }

        public bool isCSList { get; private set; }
        public int highestVARNumber { get; private set; }

        public bool isLocal { get; private set; }

        private Dictionary<int, bool> varVersionsCSListRef;
        private Dictionary<int, HashSet<string>> varVersionsPathVariant;
        private Dictionary<string, bool> pathVariantConcealed;
        private Dictionary<string, bool> pathVariantPreferred;

        private HashSet<string> pathVariantsInSameVAR;

        public HashSet<string> varVersionsNoCSListRef { get; private set; }

        public PluginTagSet pluginTagSet { get; private set; }

        public bool anyVariantsPreferred {
            get
            {
                foreach (KeyValuePair<string,bool> kvp in pathVariantPreferred)
                {                    
                    if (kvp.Value) return true;
                }
                return (false);
            }
        }
        public bool allVariantsConcealed
        {
            get
            {
                foreach (KeyValuePair<string, bool> kvp in pathVariantConcealed)
                {
                    if (!kvp.Value) return false;
                }
                return (true);
            }
        }

        public void SetVariantConcealed (string path, bool value, bool store =true)
        {
            if (pathVariantConcealed.ContainsKey(path))
            {
                if (pathVariantsInSameVAR.Contains(path)) pathVariantConcealed[path] = value;
                else
                {
                    foreach (string pvcKey in pathVariantConcealed.Keys.ToList())
                    {
                        if (!pathVariantsInSameVAR.Contains(path)) pathVariantConcealed[pvcKey] = value;
                    }
                }
                if (store) StorePluginVariantMetaData(path);
            }
            else Utils.LogDebug("SetVariantConcealed: path variant '" + path + "' is not in plugin manifest");
        }
        public void SetAllVariantConcealedExcept(string path)
        {
            HashSet<string> keys ;
            if (pathVariantsInSameVAR.Contains(path)) keys = new HashSet<string>(pathVariantConcealed.Keys);
            else keys = new HashSet<string>(pathVariantsInSameVAR);

            foreach (string key in keys)
            {
                if (key != path) SetVariantConcealed(key, true);
            }
        }

        public bool GetVariantConcelealed(string path)
        {
            if (pathVariantConcealed.ContainsKey(path)) return (pathVariantConcealed[path]);
//            else if (pathVariantConcealed.Count == 1) return pathVariantConcealed.First().Value;
            else Utils.LogDebug("GetVariantConcealed: path variant '" + path + "' is not in plugin manifest");
            return false;
        }
        public void SetVariantPreferred(string path, bool value, bool store = true)
        {
            if (pathVariantPreferred.ContainsKey(path))
            {
                if (pathVariantsInSameVAR.Contains(path)) pathVariantPreferred[path] = value;
                else
                {
                    foreach (string pvcKey in pathVariantPreferred.Keys.ToList())
                    {
                        if (!pathVariantsInSameVAR.Contains(path)) pathVariantPreferred[pvcKey] = value;
                    }
                }
                if (store) StorePluginVariantMetaData(path);
            }
            else
            {
                Utils.LogDebug("SetVariantPreferred: path variant '" + path + "' is not in plugin manifest");
            }
        }
        public bool GetVariantPreferred(string path)
        {
            if (pathVariantPreferred.ContainsKey(path)) return (pathVariantPreferred[path]);
            else Utils.LogDebug("GetVariantPreferred: path variant '" + path + "' is not in plugin manifest");

            return false;
        }
        private int _pluginType=PluginAtomTypes.General;
        public int pluginType {
            get
            {
                return (_pluginType);
            }
            set
            {
                _pluginType = value;
                StorePluginMetaData();
            }
        }

        public string pluginNameNoFileExt
        {
            get
            {
                if (pluginName.EndsWith(".cs")) return pluginName.Substring(0, pluginName.Length - 3);
                if (pluginName.EndsWith(".dll")) return pluginName.Substring(0, pluginName.Length - 4);
                if (pluginName.EndsWith(".cslist")) return pluginName.Substring(0, pluginName.Length - 7);
                return (pluginName);
            }
        }

        public int varVersionCount { get
            {
                if (isLocal) return 0;
                return (varVersionsNoCSListRef.Count);
            }
        }

        public bool referencedByCSList
        {
            get
            {
                return (varVersionsCSListRef[highestVARNumber]);
            }
        }

        public string pluginManifestEntryName
        {
            get {
                if (isLocal) return (creator + "." + pluginName);
                return (creator + "." + varPackageName + "." + pluginName);
            }
        }
        public string pmeVariantFolderName
        {
            get
            {
                if (isLocal) return (creator + "." + pluginName);
                return (creator + "." + varPackageName + "." + pluginName);
            }
        }


        public PluginManifestEntry(string _creator, string _path, string _pluginName, string _varPackageName, int _varVersion, TagsManifest tagManifest, bool _csList = false)
        {          
            creator = _creator;
            isLocal = false;
            varPackageName = _varPackageName;
            CommonContructor(_pluginName, tagManifest, _csList);

            AddVARVersion(_varVersion, _path);
            RestorePluginMetaData(PluginAtomTypes.General);
        }
        public PluginManifestEntry(string _path, string _pluginName, TagsManifest tagManifest, bool _csList = false)
        {           
            creator = "LOCAL";
            isLocal = true;
            varPackageName = "";
            CommonContructor(_pluginName, tagManifest, _csList);

            AddVARVersion(0, _path);
            RestorePluginMetaData(PluginAtomTypes.General);
        }
        private void CommonContructor(string _pluginName, TagsManifest tagManifest, bool _csList)
        {
            varVersionsCSListRef = new Dictionary<int, bool>();
            varVersionsPathVariant = new Dictionary<int, HashSet<string>>();
            varVersionsNoCSListRef = new HashSet<string>();
            pathVariantConcealed = new Dictionary<string, bool>();
            pathVariantPreferred = new Dictionary<string, bool>();
            pathVariantsInSameVAR = new HashSet<string>();
            pluginName = _pluginName;
            isCSList = _csList;
            highestVARNumber = -1;
            pluginTagSet = new PluginTagSet(tagManifest, this);
        }

        public bool CheckCreatorMatchesFilter(JSONStorableCreatorChooser creatorJSON)
        {
            if (creatorJSON.isValFixedIndex)
            {
                if (creatorJSON.displayVal == "ALL") return (true);
                else if (creatorJSON.displayVal == "VAR ONLY") return (!isLocal);
                else if (creatorJSON.displayVal == "LOCAL ONLY") return (isLocal);
            }
            return (creator == creatorJSON.displayVal);
        }
        public void StorePluginMetaData()
        {
            JSONClass metaData = new JSONClass();
            metaData["PAversion"] = Const.PAversion;
            metaData["PAversionFloat"].AsFloat = Const.PAversionFloat;
            metaData["metaDataStoreFormat"].AsInt = Const.MetaDataStoreFormat;
            metaData["pluginName"] = pluginManifestEntryName;

            metaData["pluginType"] = PluginAtomTypes.Label(pluginType);
            metaData["tags"] = pluginTagSet.GetActiveTagsJSON();
   
            if (isLocal)
            {
                FileManagerSecure.CreateDirectory(Const.LocalPluginMetaDataFolder + "\\" + varVersionsPathVariant[0].First());
                SuperController.singleton.SaveJSON(metaData, Const.LocalPluginMetaDataFolder + "\\" + varVersionsPathVariant[0].First() + "\\" + pluginManifestEntryName + Const.PluginMetaDataFileExt);
            }
            else SuperController.singleton.SaveJSON(metaData, Const.VARPluginMetaDataFolder + "\\" + pluginManifestEntryName + Const.PluginMetaDataFileExt);
        }
        private void StorePluginVariantMetaData(string pathVariant)
        {
            JSONClass metaData = new JSONClass();
            metaData["PAversion"] = Const.PAversion;
            metaData["PAversionFloat"].AsFloat = Const.PAversionFloat;
            metaData["metaDataStoreFormat"].AsInt = Const.MetaDataStoreFormat;
            metaData["pluginName"] = pluginManifestEntryName;

            metaData["concealed"].AsBool = GetVariantConcelealed(pathVariant);
            metaData["preferred"].AsBool = GetVariantPreferred(pathVariant);

            string folderName = Const.VARPluginMetaDataFolder + "\\" + pmeVariantFolderName + "\\" + pathVariant.Replace('/','\\') ;
            if (isLocal) folderName = Const.LocalPluginMetaDataFolder + "\\" + pathVariant.Replace('/', '\\');

            FileManagerSecure.CreateDirectory(folderName);
            SuperController.singleton.SaveJSON(metaData, folderName + "\\" + pmeVariantFolderName + Const.PluginFolderVariantMetaDataFileExt);
        }


        private void RestorePluginVariantMetaData(string pathVariant)
        {
            string fileName = Const.VARPluginMetaDataFolder + "\\" + pmeVariantFolderName + "\\" + pathVariant.Replace('/', '\\') + "\\" + pmeVariantFolderName + Const.PluginFolderVariantMetaDataFileExt;
            if (isLocal) fileName = Const.LocalPluginMetaDataFolder + "\\" + pathVariant.Replace('/', '\\') + "\\" + pmeVariantFolderName + Const.PluginFolderVariantMetaDataFileExt;
            if (FileManagerSecure.FileExists(fileName))
            {
                JSONClass metaData = (JSONClass)SuperController.singleton.LoadJSON(fileName);
                if (metaData["metaDataStoreFormat"] != null && metaData["metaDataStoreFormat"].AsInt > Const.MetaDataStoreFormat) Utils.LogDebug("<RestorePluginVariantMetaData> Plugin Variant Metadata Store Format for '"+ fileName+"' is a higher version than supported.");

                if (metaData["concealed"] != null) SetVariantConcealed(pathVariant, metaData["concealed"].AsBool,false);
                if (metaData["preferred"] != null) SetVariantPreferred(pathVariant, metaData["preferred"].AsBool,false);
            }
        }

        private void RestorePluginMetaData(int defaultPluginType)
        {
            string fileName = Const.VARPluginMetaDataFolder + "\\" + pluginManifestEntryName + Const.PluginMetaDataFileExt;
            if (isLocal) fileName = Const.LocalPluginMetaDataFolder + "\\" + varVersionsPathVariant[0].First() + "\\" + pluginManifestEntryName + Const.PluginMetaDataFileExt;
            if (FileManagerSecure.FileExists(fileName))
            {
                JSONClass metaData = (JSONClass)SuperController.singleton.LoadJSON(fileName);
                if (metaData["metaDataStoreFormat"] != null && metaData["metaDataStoreFormat"].AsInt > Const.MetaDataStoreFormat) Utils.LogDebug("<RestorePluginMetaData> Plugin Metadata Store Format for '" + fileName + "' is a higher version than supported.");

                    if (metaData["pluginType"] != null) _pluginType = PluginAtomTypes.Enumerate(metaData["pluginType"]);
                else _pluginType = defaultPluginType;

                if (metaData["tags"] != null)
                {
                    pluginTagSet.SetActiveTags((JSONArray)metaData["tags"]);
                }
            }
            else _pluginType = defaultPluginType;
        }
        public int GetPathVariantCount(int varVersion)
        {
            return (varVersionsPathVariant[varVersion].Count);
        }
        public int GetConcealedPathVariantCount()
        {
            return (pathVariantConcealed.Values.Where(p => p == true).Count());
        }
        public int GetPathVariantsInSameVARCount()
        {
            return (pathVariantsInSameVAR.Count());
        }
        public int GetPackageVersionsCount()
        {
            return (varVersionsNoCSListRef.Count);
        }

        public void AddVARVersion(int version, string path)
        {
            if (!varVersionsCSListRef.ContainsKey(version))
            {
                varVersionsCSListRef.Add(version, false);
                varVersionsNoCSListRef.Add(version.ToString());
            }

            if (!varVersionsPathVariant.ContainsKey(version))
            {
                varVersionsPathVariant.Add(version, new HashSet<string>());
                varVersionsPathVariant[version].Add(path);
            }
            else if (!varVersionsPathVariant[version].Contains(path)) varVersionsPathVariant[version].Add(path);

            if (!pathVariantConcealed.ContainsKey(path)) pathVariantConcealed.Add(path, false);
            if (!pathVariantPreferred.ContainsKey(path)) pathVariantPreferred.Add(path, false);

            if (varVersionsPathVariant[version].Count == 2)
            {
                foreach (string versionPath in varVersionsPathVariant[version]) pathVariantsInSameVAR.Add(versionPath);              
            }
            else if (varVersionsPathVariant[version].Count > 2) pathVariantsInSameVAR.Add(path);

            RestorePluginVariantMetaData(path);

            if (version > highestVARNumber) highestVARNumber = version;
        }
        public void SetVARVersionReferencedInCSList(int version)
        {
            if (varVersionsCSListRef.ContainsKey(version))
            {
                varVersionsCSListRef[version] = true;
                varVersionsNoCSListRef.Remove(version.ToString());
            }
        }

        public List<string> GetPluginsVersionsList()
        {
            List<int> intVerList = varVersionsCSListRef.Keys.ToList();
            intVerList.Sort();
            intVerList.Reverse();
            List<string> stringVerList = intVerList.ConvertAll(Convert.ToString);
            return(stringVerList);
            
        }
        public List<string> GetPathVariantsList(int varVersion, bool stripCustomScripts = false, int preferredFilter = PreferredFilter.any, bool showConcealed = false)
        {
            List<string> pathVariantsList = varVersionsPathVariant[varVersion].ToList();
            if (stripCustomScripts || preferredFilter != PreferredFilter.any || showConcealed == false)
            {
                List<string> refinedPathVariantList = new List<string>();
                foreach (string path in pathVariantsList)
                {
                    string finalPath = path;
                    bool preferred = pathVariantPreferred[path];
                    bool concealed= pathVariantConcealed[path];

                    if (path.StartsWith("Custom/Scripts") && stripCustomScripts) finalPath=path.Substring(14);
                    if ((preferredFilter== PreferredFilter.any || (preferredFilter==PreferredFilter.notPreferred && !preferred)|| (preferredFilter==PreferredFilter.preferred && preferred)) &&
                        (showConcealed || !concealed)) refinedPathVariantList.Add(finalPath);
                }
                return refinedPathVariantList;
            }

            return (pathVariantsList);
        }
    }

    public class VARPackageID
    {
        public string varCreator { get; private set; }
        public string varPackageName { get; private set; }
        public int varVersion { get; private set; }
        public string varFileName { get; private set; }

        public VARPackageID(string varFilePath)
        {
            string temp = FileManagerSecure.GetFileName(varFilePath);
            varFileName = temp.Substring(0, temp.Length - 4);
            varCreator = varFileName.Substring(0, varFileName.IndexOf('.'));
            temp = varFileName.Substring(varCreator.Length + 1);
            varPackageName = temp.Substring(0, temp.IndexOf('.'));
            temp = varFileName.Substring(varCreator.Length + varPackageName.Length + 2);
            varVersion = int.Parse(temp);
        }
    }

}
